/* $Id: procstream.h,v 1.4 2005/12/09 05:07:08 yusuke Exp $
 *
 *    Copyright (c) 2005, Yusuke Miyao
 *
 *    You may distribute this file under the terms of the Artistic License.
 */

#ifndef __procstream_h
#define __procstream_h

#include "lconfig.h"
#include "profile.h"
#include <iostream>
#include <termios.h>

#ifdef HAVE_SYS_TYPES_H
#  include <sys/types.h>
#endif // HAVE_SYS_TYPES_H

#ifdef HAVE_SYS_WAIT_H
#  include <sys/wait.h>
#endif // HAVE_SYS_WAIT_H

#ifdef HAVE_STROPTS_H
#  include <stropts.h>
#endif // HAVE_STROPTS_H

#ifdef HAVE_FCNTL_H
#  include <fcntl.h>
#endif // HAVE_FCNTL_H

#ifdef HAVE_SYS_IOCTL_H
#  include <sys/ioctl.h>
#endif // HAVE_SYS_IOCTL_H

#ifdef HAVE_PTY_H
#  include <pty.h>
#endif // HAVE_PTH_H

namespace lilfes {

//////////////////////////////////////////////////////////////////////
////
//// streambuf for process communication
////
//////////////////////////////////////////////////////////////////////

class process_stream_buf : public std::streambuf {
public:
  typedef size_t size_type;
  typedef char char_type;
  static const size_type DEFAULT_BUFFER_SIZE = 4096;

private:
  size_type buffer_size;
  char_type* buffer;

  int fd;
  int pid;
  //char* command;
  //int argc;
  //char** argv;
  bool read_mode;
  bool is_eof;

public:
  process_stream_buf( size_type s = DEFAULT_BUFFER_SIZE ) {
    buffer_size = s;
    buffer = new char_type[ buffer_size ];
  }
  virtual ~process_stream_buf() {
    close();
    delete [] buffer;
  }

  int get_fd() const { return fd; }
  int get_pid() const { return pid; }

  virtual bool is_open() const {
    return pid > 0;
  }
  //virtual bool eof() const { return is_eof; }

  virtual process_stream_buf* close() {
    if ( pid > 0 ) {
      if ( ::close( fd ) < 0 || waitpid( pid, NULL, 0 ) < 0 ) {
        RUNERR( "Cannot close process stream" );
        pid = 0;
        return NULL;
      }
    }
    pid = 0;
    return this;
  }

  virtual process_stream_buf* open(char* command_, int, char** argv_, bool read_mode_) {
    //command =  command_;
    //argc = argc_;
    //argv = argv_;
    pid = 0;
    is_eof = false;
    read_mode = read_mode_;
//     read_mode = false;
//     if ( strcmp( option_, "r" ) == 0 ) {
//       read_mode = true;
//     } else if ( strcmp( option_, "w" ) == 0 ) {
//       read_mode = false;
//     } else {
//       RUNERR( "Invalid open mode for process stream: " << option_ );
//       return NULL;
//     }
    int pipe_fd[ 2 ];
    if ( pipe( pipe_fd ) < 0 ) {
      RUNERR( "Cannot create a new pipe for process stream" );
      return NULL;
    }
    pid = fork();
    if ( pid < 0 ) {
      RUNERR( "Cannot fork a new process" );
      pid = 0;
      return NULL;
    }
    if ( pid == 0 ) {
      if ( read_mode ) {
        dup2( pipe_fd[ 1 ], 1 );
        // stderr is ignored
        //dup2( pipe_fd[ 1 ], 2 );
      } else {
        dup2( pipe_fd[ 0 ], 0 );
      }
      ::close( pipe_fd[ 0 ] );
      ::close( pipe_fd[ 1 ] );
      execvp( command_, argv_ );
      RUNERR( "Execution of command failed: " << command_ );
      exit( 1 );
    }
    if ( read_mode ) {
      fd = pipe_fd[ 0 ];
      ::close( pipe_fd[ 1 ] );
    } else {
      fd = pipe_fd[ 1 ];
      ::close( pipe_fd[ 0 ] );
    }
    return this;
  }

  int sync() {
    if ( ! read_mode ) {
      char_type* buffer_ptr = buffer;
      while ( buffer_ptr < pptr() ) {
        int count = write( fd, buffer, pptr() - buffer );
        if ( count <= 0 ) return -1;
        buffer_ptr += count;
      }
      setp( buffer, buffer + buffer_size );
    }
    return 0;
  }

  int underflow() {
    if ( ! read_mode ) return EOF;
    int count = read(fd, buffer, buffer_size);
    setg( buffer, buffer, buffer + count );
    return count <= 0 ? EOF : *buffer;
  }

  int overflow( int c = EOF ) {
    if ( read_mode ) return EOF;
    sync();
    if ( c == EOF ) return EOF;
    *buffer = c;
    pbump( 1 );
    return c;
  }
};

//////////////////////////////////////////////////////////////////////
////
//// streambuf for pty communication
////
//////////////////////////////////////////////////////////////////////

class pty_stream_buf : public std::streambuf {
public:
  typedef size_t size_type;
  typedef char char_type;
  static const size_type DEFAULT_BUFFER_SIZE = 4096;

private:
  size_type buffer_size;
  char_type* buffer;

  int fd;
  int pid;
  //char* command;
  //int argc;
  //char** argv;
  bool is_eof;

public:
  pty_stream_buf( size_type s = DEFAULT_BUFFER_SIZE ) {
    buffer_size = s;
    buffer = new char_type[ buffer_size ];
  }
  virtual ~pty_stream_buf() {
    close();
    delete [] buffer;
  }

  int get_fd() const { return fd; }
  int get_pid() const { return pid; }

  virtual bool is_open() const {
    return pid > 0;
  }
  //virtual bool eof() const { return is_eof; }

  virtual pty_stream_buf* close() {
    if ( pid > 0 ) {
      if ( ::close( fd ) < 0 || waitpid( pid, NULL, 0 ) < 0 ) {
        RUNERR( "Cannot close pty stream" );
        pid = 0;
        return NULL;
      }
    }
    is_eof = true;
    pid = 0;
    return this;
  }

  virtual pty_stream_buf* open(char* command_, int, char** argv_ ) {
    //command = command_;
    //argc = argc_;
    //argv = argv_;
    pid = 0;
    is_eof = false;
#if ( ( defined( HAVE_LIBUTIL ) && defined( HAVE_OPENPTY ) ) || defined( HAVE_GRANTPT ) )
    int slave_fd = 0;
#if ( defined( HAVE_LIBUTIL ) && defined( HAVE_OPENPTY ) )
    if ( openpty( &fd, &slave_fd, NULL, NULL, NULL ) < 0 ) {
      RUNERR( "Cannot open pseudo tty" );
      return NULL;
    }
#else // HAVE_OPENPTY
    fd = ::open( "/dev/ptmx", O_RDWR | O_NOCTTY );
    if ( fd < 0 ) {
      RUNERR( "Cannot open pseudo tty" );
      return NULL;
    }
    if ( grantpt( fd ) < 0 || unlockpt( fd ) < 0 ) {
      RUNERR( "Cannot initialize pseudo tty" );
      return NULL;
    }
    char* slave_name = ptsname( fd );
    if ( slave_name == NULL ) {
      RUNERR( "Cannot get the name of a slave tty" );
      return NULL;
    }
    slave_fd = ::open( slave_name, O_RDWR | O_NOCTTY );
    if ( slave_fd < 0 ) {
      RUNERR( "Cannot open a slave tty" );
      return NULL;
    }
#ifdef I_PUSH
    ioctl( slave_fd, I_PUSH, "ptem" );
    ioctl( slave_fd, I_PUSH, "ldterm" );
    ioctl( slave_fd, I_PUSH, "ttcompat" );
#endif // I_PUSH
#endif // HAVE_OPENPTY
    struct termios stermios;
    if ( tcgetattr( slave_fd, &stermios ) < 0 ) {
      RUNERR( "Cannot get terminal attributes" );
      return NULL;
    }
    stermios.c_lflag &= ~( ECHO | ECHOE | ECHOK | ECHONL );
    stermios.c_lflag |= ICANON;
    stermios.c_oflag &= ~( ONLCR );
    if ( tcsetattr( slave_fd, TCSANOW, &stermios ) < 0 ) {
      RUNERR( "Cannot set terminal attributes" );
      return NULL;
    }
    pid = fork();
    if ( pid < 0 ) {
      RUNERR( "Cannot fork a new process" );
      pid = 0;
      return NULL;
    }
    if ( pid == 0 ) {
      dup2( slave_fd, 0 );
      dup2( slave_fd, 1 );
      // stderr is ignored
      //dup2( slave_fd, 2 );
      ::close( fd );
      ::close( slave_fd );
      execvp( command_, argv_ );
      RUNERR( "Execution of command failed: " << command_ );
      exit( 1 );
    }
    ::close( slave_fd );
    return this;
#else // HAVE_OPENPTY || HAVE_GRANTPT
    fd = -1;
    RUNERR( "This executable does not support pseudo-tty stream" );
    return NULL;
#endif // HAVE_OPENPTY || HAVE_GRANTPT
  }

  int sync() {
    char_type* buffer_ptr = buffer;
    while ( buffer_ptr < pptr() ) {
      int count = write( fd, buffer, pptr() - buffer );
      if ( count <= 0 ) return -1;
      buffer_ptr += count;
    }
    setp( buffer, buffer + buffer_size );
    return 0;
  }

  int underflow() {
    int count = read(fd, buffer, buffer_size);
    setg( buffer, buffer, buffer + count );
    return count <= 0 ? EOF : *buffer;
  }

  int overflow( int c = EOF ) {
    sync();
    if ( c == EOF ) return EOF;
    *buffer = c;
    pbump( 1 );
    return c;
  }
};

//////////////////////////////////////////////////////////////////////
////
////  iostream interface with process stream
////
//////////////////////////////////////////////////////////////////////

class process_istream : public std::istream {
private:
  process_stream_buf bf_buf;

public:
  process_istream() : std::istream( NULL ), bf_buf() {
    init( &bf_buf );
  }
  explicit process_istream( char* command_, int argc_, char** argv_ )
    : std::istream( NULL ), bf_buf() {
    init( &bf_buf );
    open( command_, argc_, argv_ );
  }
  virtual ~process_istream() {
  }

public:
  int get_fd() const { return bf_buf.get_fd(); }
  int get_pid() const { return bf_buf.get_pid(); }

  void open( char* command_, int argc_, char** argv_ ) {
    if ( ! bf_buf.open( command_, argc_, argv_, true ) ) {
      setstate( std::ios::failbit );
    }
  }
  void close() {
    if ( ! bf_buf.close() ) {
      setstate( std::ios::failbit );
    }
  }
};

class process_ostream : public std::ostream {
private:
  process_stream_buf bf_buf;

public:
  process_ostream() : std::ostream( NULL ), bf_buf() {
    init( &bf_buf );
  }
  explicit process_ostream( char* command_, int argc_, char** argv_ )
    : std::ostream( NULL ), bf_buf() {
    init( &bf_buf );
    open( command_, argc_, argv_ );
  }
  virtual ~process_ostream() {
  }

public:
  int get_fd() const { return bf_buf.get_fd(); }
  int get_pid() const { return bf_buf.get_pid(); }

  void open( char* command_, int argc_, char** argv_ ) {
    if ( ! bf_buf.open( command_, argc_, argv_, false ) ) {
      setstate( std::ios::failbit );
    }
  }
  void close() {
    if ( ! bf_buf.close() ) {
      setstate( std::ios::failbit );
    }
  }
};

//////////////////////////////////////////////////////////////////////
////
////  iostream interface with pty stream
////
//////////////////////////////////////////////////////////////////////

class pty_stream : public std::iostream {
private:
  pty_stream_buf bf_buf;

public:
  pty_stream() : std::iostream( NULL ), bf_buf() {
    init( &bf_buf );
  }
  pty_stream( char* command_, int argc_, char** argv_ )
    : std::iostream( NULL ), bf_buf() {
    init( &bf_buf );
    open( command_, argc_, argv_ );
  }
  virtual ~pty_stream() {
  }

public:
  int get_fd() const { return bf_buf.get_fd(); }
  int get_pid() const { return bf_buf.get_pid(); }

  void open( char* command_, int argc_, char** argv_ ) {
    if ( ! bf_buf.open( command_, argc_, argv_ ) ) {
      setstate( std::ios::failbit );
    }
  }
  void close() {
    if ( ! bf_buf.close() ) {
      setstate( std::ios::failbit );
    }
  }
};

} // namespace lilfes

#endif // __procstream_h
