/* $Id: sockstream.h,v 1.2 2005/05/18 10:29:10 ninomi Exp $
 *
 *    Copyright (c) 2005, Yusuke Miyao
 *
 *    You may distribute this file under the terms of the Artistic License.
 */

#ifndef __sockstream_h
#define __sockstream_h

#include "lconfig.h"
#include "profile.h"
#include <iostream>
#include <sys/types.h>
#include <netdb.h>

#include <cerrno>

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif // HAVE_UNISTD_H

#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#include <arpa/inet.h>
#else // HAVE_SYS_SOCKET_H
#ifdef HAVE_WINSOCK_H
#include <winsock.h>
#endif // HAVE_WINSOCK_H
#endif // HAVE_SYS_SOCKET_H

namespace lilfes {

//////////////////////////////////////////////////////////////////////
////
////  streambuf for socket connection
////
//////////////////////////////////////////////////////////////////////

class socket_stream_buf_base : public std::streambuf {
public:
  typedef size_t size_type;
  typedef char char_type;
  static const size_type DEFAULT_BUFFER_SIZE = 4096;

private:
  size_type buffer_size;
  char_type* buffer;

protected:
  int sock;

public:
  explicit socket_stream_buf_base( int sock_ = 0, size_type s = DEFAULT_BUFFER_SIZE ) : sock( sock_ ) {
    buffer_size = s;
    buffer = new char_type[ buffer_size ];
  }

  virtual ~socket_stream_buf_base() {
    close();
    delete [] buffer;
  }

  virtual bool is_open() const {
    return sock != 0;
  }

  int socket_id() const { return sock; }

  virtual socket_stream_buf_base* close() {
    int ret = sock == 0 ? 0 : ::close(sock);
    sock = 0;
    return ret == 0 ? this : NULL;
  }

  int sync() {
    char_type* buffer_ptr = buffer;
    while ( buffer_ptr < pptr() ) {
      int count = send( sock, buffer, pptr() - buffer, 0 );
      if ( count <= 0 ) return -1;
      buffer_ptr += count;
    }
    setp( buffer, buffer + buffer_size );
    return 0;
  }

  int underflow() {
    int count = recv(sock, buffer, buffer_size, 0);
    setg( buffer, buffer, buffer + count );
    return count <= 0 ? EOF : *buffer;
  }

  int overflow( int c = EOF ) {
    sync();
    if ( c == EOF ) return EOF;
    *buffer = c;
    pbump( 1 );
    return c;
  }
};

//////////////////////////////////////////////////////////////////////
////
////  client socket
////
//////////////////////////////////////////////////////////////////////

class socket_stream_buf : public socket_stream_buf_base {
private:
  //static const struct protoent* TCP_PROTO = getprotobyname("tcp");

public:
  socket_stream_buf() : socket_stream_buf_base( 0 ) {
  }

  virtual ~socket_stream_buf() {
    if ( is_open() ) close();
  }

  virtual socket_stream_buf* open( const std::string& hostname_, int port_number_ ) {
    struct sockaddr_in sai;
    struct hostent *he;
    if( ( sock = socket( AF_INET, SOCK_STREAM, 0 ) ) < 0 ) { // $B%=%1%C%H$r:n$k(B
      RUNERR("Can't open a socket");
      sock = 0;
      return NULL;
    }
    //   int option = 1;   // NODELAY $B$rM-8z$K$9$k(B
    //   if( ( setsockopt( sock, TCP_PROTO->p_proto,
    //                     0, (char *)&option, sizeof(option) ) ) < 0 ) {
    //     // TCP $B$NCY1dAw?.$r6X;_(B
    //     RUNERR("Can't set an option of the socket");
    //     sock = 0;
    //     return;
    //   }
    sai.sin_family = AF_INET;  // $B%$%s%?!<%M%C%H$N%]!<%H(B
    sai.sin_port = htons(port_number_ - 2);  // $B$3$N%]!<%HHV9f$O;H$o$l$J$$$O$:!)(B
    sai.sin_addr.s_addr = INADDR_ANY;// $B$I$N%k!<%H$+$i$G$b$$$$(B
    memset(sai.sin_zero,0,8);
    sai.sin_family = AF_INET;// $B@\B3@h$O%$%s%?!<%M%C%H(B
    sai.sin_port = htons(port_number_);// $B@\B3@h$N%]!<%HHV9f(B
    he = gethostbyname(hostname_.c_str());// $B%[%9%HL>$+$i%"%I%l%9$r2r7h(B
    if( he == NULL ) {  // $B%(%i!<(B
      RUNERR("Can't find the host: " << hostname_ );
      sock = 0;
      return NULL;
    }
    sai.sin_addr.s_addr = *(long *)he->h_addr_list[0];
    // $BJ#?t%"%I%l%9$,$"$k>l9g$b$"$k$N$G!":G=i$N#1%3$r;H$&(B
    //printf("host address: %d.%d.%d.%d\n",(unsigned char)he->h_addr_list[0][0],(unsigned char)he->h_addr_list[0][1],(unsigned char)he->h_addr_list[0][2],(unsigned char)he->h_addr_list[0][3]);
    if( connect(sock, (sockaddr *)&sai, sizeof(sai)) == -1 ) {// $B@\B3$9$k!#(B
      RUNERR("Can't connect to the socket: " << hostname_ << " (" << port_number_ << ")" );
      sock = 0;
      return NULL;
    }
    //printf("connected sock =%d \n",sock);
    return this;
  }
};

//////////////////////////////////////////////////////////////////////
////
////  server stream
////
//////////////////////////////////////////////////////////////////////

class server_stream_buf : public socket_stream_buf_base {
private:
  int server_sock;
  struct sockaddr_in server, client;

public:
  server_stream_buf() : socket_stream_buf_base( 0 ), server_sock( 0 ) {
  }
  virtual ~server_stream_buf() {
    if ( is_open() ) close();
  }

  bool is_open() {
    return server_sock != 0;
  }

  server_stream_buf* close() {
    socket_stream_buf_base::close();
    if ( server_sock == 0 ) return NULL;
    int ret = ::close( server_sock );
    server_sock = 0;
    return ret == 0 ? this : NULL;
  }

  virtual bool is_connected() {
    return sock != 0;
  }

  virtual server_stream_buf* open( int port_number_ ) {
    sock = 0;
    server_sock = socket(AF_INET, SOCK_STREAM, 0);
    if (server_sock < 0) {
      RUNERR("Can't open a socket: " << strerror( errno ) );
      server_sock = 0;
      return NULL;
    }

    int sockopt_val = 1;
    if ( setsockopt( server_sock, SOL_SOCKET, SO_REUSEADDR,
                     &sockopt_val, sizeof( sockopt_val ) ) < 0 ) {
      RUNERR("Can't set socket option: " << strerror( errno ) );
      server_sock = 0;
      return NULL;
    }

    server.sin_family = AF_INET;
    server.sin_port = htons(port_number_);
    server.sin_addr.s_addr = INADDR_ANY;

    if (bind(server_sock, (struct sockaddr *)&server, sizeof(server)) < 0) {
      RUNERR("Bind: " << strerror( errno ) );
      close();
      server_sock = 0;
      return NULL;
    }

    if (listen(server_sock, 5) < 0) {
      RUNERR("Listen: " << strerror( errno ) );
      server_sock = 0;
      return NULL;
    }
    return this;
  }

  virtual int accept_connection() {
    if ( sock != 0 ) {
      ::close(sock);
    }
    sock = 0;

    socklen_t mlen = sizeof(client);
    sock = ::accept(server_sock, (struct sockaddr *)&client,
                    &mlen);
    if (sock < 0) {
      RUNERR("Accept: " << strerror( errno ) );
      sock = 0;
    }
    return sock;
  }

  virtual void close_connection() {
    if ( sock != 0 ) {
      ::close( sock );
    }
    sock = 0;
  }
};

//////////////////////////////////////////////////////////////////////
////
////  iostream interface with socket stream
////
//////////////////////////////////////////////////////////////////////

class socket_stream : public std::iostream {
private:
  socket_stream_buf bf_buf;

public:
  socket_stream() : std::iostream( NULL ), bf_buf() {
    init( &bf_buf );
  }
  explicit socket_stream( const std::string& host, int port )
    : std::iostream( NULL ), bf_buf() {
    init( &bf_buf );
    open( host, port );
  }
  virtual ~socket_stream() {
  }

public:
  void open( const std::string& host, int port ) {
    if ( ! bf_buf.open( host, port ) ) {
      setstate( std::ios::failbit );
    }
  }
  void close() {
    if ( ! bf_buf.close() ) {
      setstate( std::ios::failbit );
    }
  }
};

//////////////////////////////////////////////////////////////////////
////
////  iostream interface with server stream
////
//////////////////////////////////////////////////////////////////////

class server_stream : public std::iostream {
private:
  server_stream_buf bf_buf;

public:
  server_stream() : std::iostream( NULL ), bf_buf() {
    init( &bf_buf );
  }
  explicit server_stream( int port )
    : std::iostream( NULL ), bf_buf() {
    init( &bf_buf );
    open( port );
  }
  virtual ~server_stream() {
  }

public:
  void open( int port ) {
    if ( ! bf_buf.open( port ) ) {
      setstate( std::ios::failbit );
    }
  }
  void close() {
    if ( ! bf_buf.close() ) {
      setstate( std::ios::failbit );
    }
  }

  bool is_connected() {
    return bf_buf.is_connected();
  }
  int accept_connection() {
    return bf_buf.accept_connection();
  }
  void close_connection() {
    bf_buf.close_connection();
  }

};

} // namespace lilfes

#endif // __sockstream_h
