/* $Id: machine2.cpp,v 1.4 2004/10/16 11:10:05 yusuke Exp $ 
 *
 *    Copyright (c) 1997-1998, Makino Takaki
 *
 *    You may distribute this file under the terms of the Artistic License.
 *
 */

#include "machine.h"
#include "proc.h"
#include "structur.h"
#include "codelist.h"
#include "builtin.h"
#include <sstream>
#include <iomanip>

namespace lilfes {

static const char rcsid[] = "$Id: machine2.cpp,v 1.4 2004/10/16 11:10:05 yusuke Exp $";

machine::machine()
	: local("local", 1024)
{
	for( int i=0; i<NHEAP-1; i++ )
	{
		std::ostringstream oss;
		oss << "heap " << i;
		heap[i].SetName(oss.str().c_str());
	}
	heap[NHEAP-1].SetName("stack");
	trail.SetName("trail");
	
	
	HB = MakeCoreP(1, 0);
	HP = MakeCoreP(1, 1);
	HA = CORE_P_INVALID;
	HU1 = CORE_P_INVALID;
	HU2 = CORE_P_INVALID;
	SP = int2cp(STACK_BASE);
	FP = int2cp(0);
	RP = int2cp(0);
#ifdef TRAIL
	TP = int2cp(0);
	OP = CORE_P_INVALID;
	B0 = CORE_P_INVALID;
#endif
	IP = NULL;
	CIP = NULL;
	for( int i=0; i<1024; i++ )
		local[i] = VAR2c(bot);
	CheckHeapLimit(HP);
	ClearHeap(0);
	WriteHeap(MakeCoreP(1, 0), PTR2c(HP));
	WriteHeap(MakeCoreP(2, 0), PTR2c(MakeCoreP(2, 1)));
#ifdef N1
	n1currentslot = GetSlot(HP);
	n1currenthbslot = GetSlot(HB);
#endif
        ConstraintEnabled = true;
}

machine::~machine()
{
	
}

void machine::TraceOutput( int tt, const type *t, int x )
{
	const char *s = t->GetName();
	
	std::cout << "TRACE: "
		 << ( tt==0 ? "ENTER" :
		      tt==1 ? "SUCCESS" :
		      tt==2 ? "REENTER" :
		      tt==3 ? "FAIL" : "UNKNOWN" )
		 << " " << s;
	if( tt != 3 )
		std::cout << "(" << (x+1) << ")";
	std::cout << std::endl;
	procedure *p = procedure::Search(t);
	ASSERT( p != NULL );
	int arity = p->GetArity();

	outform of;

        for( int i=0; i<arity; i++ ) {
	    of.AddAtRight(FSP(this, c2PTR(ReadLocal(i))).DisplayAVM());
	    if( i != arity-1 )
	        of.AddAtRight(outform(", "));
        }
	std::cout << of;
}

codelist * machine::Interpret(FSPointer f)
{ 
	codelist *cl = new codelist;
	REC(*cl);
	cl->AddCode(C_ALLOC);
	cl->AddCode((ushort)1);
	cl->AddCode((ushort)15);
	cl->AddCode(C_GETLEVEL, (ushort)-1);
	codelist *c = Interpret_internal(f.GetAddress());
	cl->AddCodelist(c);
	cl->AddCode(C_DEALLOC);
	cl->AddCode(C_RETURN);
	delete c;
	return cl; 
}

codelist * machine::Interpret_internal(core_p lst) 
{
	codelist *cl = new codelist;
	const type* c;
	int i;
	core_p h = Deref(lst);

	if( !IsSTR(ReadHeap(h)) && !IsVAR(ReadHeap(h)) )
	{
		std::cout << "strange" << std::endl;
		return cl;	
	}
	c = c2STRorVAR(ReadHeap(h));
	if( c == t_comma )
	{
		codelist *l1 = Interpret_internal(Deref(lst)+1);
		codelist *l2 = Interpret_internal(Deref(lst)+2);
		cl->AddCodelist(l1);
		cl->AddCodelist(l2);
		delete l1;
		delete l2;
	}
	else if( c == t_cut )
	{
		cl->AddCode(C_DEEPCUT, (varno)-1);
	}
	else
	{
		for (i = 0; i < c->GetNFeatures(); i++) 
		{
			REC(*cl);
			cl->AddCode(C_SETREG,(varno) i,PTR2c(h+i+1));
		}
		REC(*cl);
		cl->AddCode(C_CALL,procedure::New(c,c->GetNFeatures()));
		cl->AddCode(C_CLRTEMP,(ushort)15);
	}

	return cl;
}

void machine::DisplayStatistics()
{
	std::cout << "; (Current occupation" 
	     ": " << cp2int(GetOffset(HP)) << " heap" 
	     ", " << cp2int(GetOffset(SP))  << " stack" 
	     ", " << ( cp2int(TP) ) << " trail" 
		 " cell(s)"
#if defined(DEBUG) || defined(DOASSERT) || defined(DOPASS)
		 ", " << PASSCOUNT << " instructions"
#endif 
		 ")" << std::endl;
}

outform core::OutputCore (int start, int end)
{
	outform ret;
	for( int i=start; i<end; i++)
	{
                std::ostringstream oss;
		oss << std::setw(5) << std::hex << i << std::dec;
		oss << "  " << (*this)[i];
		ret.AddAtBottom(outform(oss.str().c_str()));
	}
	return ret;
}

outform core::OutputStack (int start, int end, int fp, int op)
{
	outform ret;
	for( int i=end-1; i>=start; i--)
	{
                std::ostringstream oss;
		if( i >= op && i < op+CF_LOCAL_OFFSET )
		{
			switch( i - op )
			{
				case CF_OP_OFFSET:		oss << "OP-->"; break;
				case CF_RETP_OFFSET:	oss << " RET>"; break;
				case CF_CIP_OFFSET:		oss << " CIP>"; break;
				case CF_FP_OFFSET:		oss << " FP->"; break;
				case CF_TP_OFFSET:		oss << " TP->"; break;
				case CF_HA_OFFSET:		oss << " HA->"; break;
				case CF_HP_OFFSET:		oss << " HP->"; break;
				case CF_B0_OFFSET:		oss << " B0->"; break;
				case CF_RP_OFFSET:		oss << " RP->"; break;
				default:                oss << "     "; break;
			}
		}
		else if( i == fp )
			oss << "FP-->";
		else
			oss << std::setw(5) << std::hex << i << std::dec;
		oss << "  " << std::setw(8) << std::hex << cell2int((*this)[i]) << std::dec;
		ret.AddAtBottom(outform(oss.str().c_str()));
	}
	return ret;
}

void machine::Dump()
{
	P1("machine::Dump()");
	std::cout << std::hex
	     <<   "HP=" << std::setw(8) << cp2int(HP) 
	     << ", HA=" << std::setw(8) << cp2int(HA)
//	     << ",HA2=" << std::setw(8) << cp2int(HA2)
	     << ",HU1=" << std::setw(8) << cp2int(HU1)
	     << ",HU2=" << std::setw(8) << cp2int(HU2)
	     << ", HB=" << std::setw(8) << cp2int(HB)
	     << std::endl
	     <<   "SP=" << std::setw(8) << cp2int(SP)
#ifdef TRAIL
	     << ", OP=" << std::setw(8) << cp2int(OP)
	     << ", TP=" << std::setw(8) << cp2int(TP)
	     << ", B0=" << std::setw(8) << cp2int(B0)
#endif
	     << ", FP=" << std::setw(8) << cp2int(FP)
	     << ", RP=" << std::setw(8) << cp2int(RP)
	     << std::endl
	     <<   "IP=" << std::setw(16) << FORCE_CAST(ptr_int, IP)
	     << ", CIP=" << std::setw(16) << FORCE_CAST(ptr_int, CIP)
	     << ", UIP=" << std::setw(16) << FORCE_CAST(ptr_int, UIP)
	     << "." << std::dec << std::endl;
	
#if 0
//	outform hpout  ( "     PREV     " );
//	hpout.AddAtBottom(heap[GetSlot(HP)-1].OutputCore(0, cp2int(1+GetOffset(heap[GetSlot(HP)-1][0]))));
	outform /*n*/hpout  ( "     HEAP     " );
	/*n*/hpout.AddAtBottom(heap[GetSlot(HP)].OutputCore(0, cp2int(GetOffset(HP))));
	outform trout  ( "     TRAIL    " );
	trout.AddAtBottom(trail.OutputCore(0, (cp2int(TP))));
	outform stout  ( "     STACK    " );
	stout.AddAtBottom(STACK.OutputStack(0, (cp2int(SP)-STACK_BASE), (cp2int(FP)-STACK_BASE), (cp2int(OP)-STACK_BASE)));
	outform reout  ( "     LOCAL    " );
	reout.AddAtBottom(local.OutputCore(0, (cp2int(RP)+8)));
	
// 	hpout.AddAtRightTop(nhpout);
 	hpout.AddAtRightTop(trout);
	hpout.AddAtRightTop(stout);
	hpout.AddAtRightTop(reout);
	std::cout << hpout;
#endif
}

//////////////////////////////////////////////////////////////////////////////
//
// PrettyPrint functions
//


static code interrupted[] = {
	C_INTERRUPTED, C_INTERRUPTED, C_INTERRUPTED, C_INTERRUPTED,
	C_INTERRUPTED, C_INTERRUPTED, C_INTERRUPTED, C_INTERRUPTED,
	C_INTERRUPTED, C_INTERRUPTED, C_INTERRUPTED, C_INTERRUPTED,
	C_INTERRUPTED, C_INTERRUPTED, C_INTERRUPTED, C_INTERRUPTED };

void machine::Interrupt()
{
	IP = interrupted;
	CIP = interrupted;
	UIP = interrupted;
	OP = CORE_P_INVALID;
	B0 = CORE_P_INVALID;

	SetTrailPoint(interrupted);
	
	SP = int2cp(STACK_BASE);

				core_p NewFP = SP+CODESIZEONSTACK;
				core_p NewSP = NewFP+1;
				CheckStackLimit(NewSP);
				
				WriteStackIP(SP,  interrupted);
				WriteStack(NewFP, PTR2c(FP));
				FP = NewFP;

	RP = int2cp(0);
	queue.FlushAll(); 
}

} // namespace lilfes
