/**
 * @module  = string
 * @cvs     = $Id: string.cpp,v 1.10 2005/06/12 09:51:06 yusuke Exp $
 * @copyright = Copyright (c) 1997-1998, Makino Takaki
 * @copyright = You may distribute this file under the terms of the Artistic License.
 * @desc    = String manipulation
 * Predicates for manipulating strings
 * @jdesc   = ʸ
 * @japanese =
 * ʸԤʤҸǤ
 * @end_japanese
*/

#include "builtin.h"
#include "utility.h"
#include <cerrno>

using namespace std;

namespace lilfes {

namespace builtin {

/**
  * @predicate	= strtolist(+STRING, -LIST) 
  * @desc	= Convert a string into a list of ASCII codes. 
  * @param	= +STRING/ string : Character string
  * @param	= -LIST/ list  : List of ASCII codes 
  * @example	= 
  > ?- strtolist("abc", X).
  X: < 97, 98, 99 > 
  * @end_example

  * @jdesc	= ʸ򡢥ɤΥꥹȤѴޤ 
  * @jparam	= +STRING/ string : ʸ
  * @jparam	= -LIST/ list : ɤΥꥹ 
  * @jexample	= 
  > ?- strtolist("abc", X).
  X: < 97, 98, 99 > 
  * @end_jexample
*/
bool strtolist( machine& m, FSP arg1, FSP arg2 )
{
	cell c = arg1.ReadCell();
	if( !IsSTG(c) )
	{
 		RUNWARN("Warning: strtolist/2 requires a string"); // Modified by mitsuisi (06/10/98)
		return false;
	}
	char *s = c2STG(c);
	while( *s )
	{
		if( arg2.Follow(hd).Unify(FSP(m, (mint)(unsigned char)*s)) == false )
		{
			return false;
		}
		arg2 = arg2.Follow(tl);
		s++;
	}
  return arg2.Coerce(nil);
}

LILFES_BUILTIN_PRED_2(strtolist, strtolist);

//////////////////////////////////////////////////////////////////////////////
// strtodlist/3 : decompose string into dlist (by mitsuisi (1999.05.20))

/**
  * @predicate	= strtodlist(+STRING, -LIST, -DLIST) 
  * @desc	= Convert a string into a list of ASCII codes and add other list to that list.
  * @param	= +STRING/ string : Character string
  * @param	= -LIST/ list  : List of ASCII codes 
  * @param	= -DLIST/ list  : Tail of the list
  * @example	= 
  * @end_example

  * @jdesc	= ʸ򡢥ɤΥꥹȤѴΥꥹȤ̤ΥꥹȤդäޤ 
  * @jparam	= +STRING/ string : ʸ
  * @jparam	= -LIST/ list : ɤΥꥹ 
  * @jparam	= -DLIST/ list : ꥹȤ
  * @jexample	= 
  * @end_jexample
*/
bool strtodlist( machine& m, FSP arg1, FSP arg2, FSP arg3 )
{
	cell c = arg1.ReadCell();
	if( !IsSTG(c) )
	{
 		RUNWARN("Warning: strtodlist/3 requires a string"); // Modified by mitsuisi (06/10/98)
		return false;
	}
	char *s = c2STG(c);
	while( *s )
	{
		if( arg2.Follow(hd).Unify(FSP(m, (mint)(unsigned char)*s)) == false )
		{
			return false;
		}
		arg2 = arg2.Follow(tl);
		s++;
	}
  return arg2.Unify(arg3);
}

LILFES_BUILTIN_PRED_3(strtodlist, strtodlist);

//////////////////////////////////////////////////////////////////////////////
// listtostr/2 : compose list into string

/**
  * @predicate	= listtostr(+LIST, -STRING) 
  * @desc	= Convert a list of ASCII codes into a string 
  * @param	= +LIST/ list  : List of ASCII codes 
  * @param	= -STRING/ string : Character string
  * @note	= If the list include a 0, the return is undefined. 
  * @example	= 
  > ?- listtostr([65,66,67], X).
  X: "ABC" 
  * @end_example

  * @jdesc	= ɤΥꥹȤʸѴޤ 
  * @jparam	= +LIST/ list : ɤΥꥹ 
  * @jparam	= -STRING/ string : ʸ
  * @jnote	= ϤΥꥹ 0 ޤޤƤưݾڤޤ 
  * @jexample	= 
  > ?- listtostr([65,66,67], X).
  X: "ABC" 
  * @end_jexample
*/

bool listtostr( machine& m, FSP arg1, FSP arg2 )
{
	char buf[8192];
	char *bufp = buf;
	uint bufidx = 0;
	uint bufsize = sizeof(buf);

	while( arg1.ReadCell() != VAR2c(nil) )
	{
		if( arg1.ReadCell() != STR2c(cons) )
		{
			RUNWARN("Warning: listtostr requires a list");
			return false;
		}
		FSP f = arg1.Follow(hd);
		if( ! IsINT(f.ReadCell()) )
		{
			RUNWARN("Warning: listtostr requires a list of integer");
			return false;
		}
		bufp[bufidx++] = c2INT(f.ReadCell());
		if( bufidx == bufsize )
		{
			int newbufsize = bufsize*3/2;
			char *newbufp = new char[newbufsize];
			memcpy(newbufp, bufp, bufsize);
			if( bufp != buf )
				delete bufp;
			bufp = newbufp;
			bufsize = newbufsize;
		}
		arg1 = arg1.Follow(tl);
	}
	bufp[bufidx] = '\0';

	FSP result(m, bufp);
	if( arg2.Unify(result) == false )
	{
          if( bufp != buf )
            delete[] bufp;

          return false;
	}
	if( bufp != buf )
		delete[] bufp;
  return true;
}

LILFES_BUILTIN_PRED_2(listtostr, listtostr);

/**
  * @predicate	= strcat(+STR1,+STR2,-RET) 
  * @desc	= Concatenate strings.
  * @param	= +STR1/ string :  string1
  * @param	= +STR2/ string :  string2
  * @param	= -RET/ string :  returned string
  * @example	= 
  * @end_example

  * @jdesc	= ʸܹ礷ޤ
  * @jparam	= +STR1/ string : ʸ1
  * @jparam	= +STR2/ string : ʸ2
  * @jparam	= +STR1/ string : ̤ʸ
  * @jexample	= 
  * @end_jexample
*/
bool strcat( machine& m, FSP arg1, FSP arg2, FSP arg3 )
{
	cell c1= arg1.ReadCell(); cell c2 = arg2.ReadCell();
	if( !IsSTG(c1) || !IsSTG(c2) )
	{
          RUNWARN("Warning: strcat/3 requires a string");
          return false;
	}

        std::string str( arg1.ReadString() );
        str += arg2.ReadString();

	return arg3.Unify( FSP( m, str.c_str() ) );
}

LILFES_BUILTIN_PRED_3(strcat, strcat);

/**
  * @predicate	= strlen(+STR,-LEN) 
  * @desc	= Length of string
  * @param	= +STR/ string :  string
  * @param	= -LEN/ string :  length of string
  * @example	= 
  * @end_example

  * @jdesc	= ʸĹ֤ޤ
  * @jparam	= +STR/ string : ʸ
  * @jparam	= +LEN/ string : Ĺ
  * @jexample	= 
  * @end_jexample
*/
bool strlen( machine& m, FSP arg1, FSP arg2 )
{
  if ( ! arg1.IsString() ) {
    RUNWARN("Warning: strlen/2 requires a string");
    return false;
  }
  char* str = arg1.ReadString();
  size_t len = ::strlen( str );
  return arg2.Unify( FSP( m, (mint)len ) );
}

LILFES_BUILTIN_PRED_2(strlen, strlen);

/**
  * @predicate	= strtotype(+STRING, -TYPE) 
  * @desc	= Convert a string into a type. 
  * @param	= +LIST/ string : String
  * @param	= -STRING/ bot : Type name 
  * @note	= In the case of the type name isn't defined, the function fails. 
  * @example	= 
  > ?- strtotype("assert/1", X).
  X: assert/1 
  * @end_example

  * @jdesc	= ʸ򷿤Ѵޤ 
  * @jparam	= +LIST/ string : ʸ
  * @jparam	= -STRING/ bot : ̾ 
  * @jnote	= ̾ƤʤΤξ硢Ԥޤ 
  * @jexample	= 
  > ?- strtotype("assert/1", X).
  X: assert/1 
  * @end_jexample
*/
bool strtotype( machine&, FSP arg1, FSP arg2 )
{
//cout << "ENTER STRTOTYPE" << endl;
//	cout << arg1.DisplayLiLFeS();
	if( !arg1.IsString() )
	{
		RUNWARN("strtotype/2 requires a string");
//		cout << arg1.DisplayLiLFeS();
		return false;
	}
	
	const type *t = module::CurrentModule()->Search(arg1.ReadString());
	
	if( t == NULL )
          return false;
	else
	{
          return arg2.Coerce(t);
	}
//cout << "LEAVE STRTOTYPE" << endl;
}

LILFES_BUILTIN_PRED_2(strtotype, strtotype);

/**
  * @predicate	= strtotype(+MODULE,+STRING, -TYPE) 
  * @desc	= Convert a string into a type in a module. 
  * @param	= +MODULE/string :module name
  * @param	= +LIST/ string : String
  * @param	= -STRING/ bot : Type name 
  * @note	= In the case of the type name isn't defined, the function fails. 
  * @example	= 
  > ?- strtotype("assert/1", X).
  X: assert/1 
  * @end_example

  * @jdesc	= ʸ⥸塼ηѴޤ 
  * @jparam	= +MODULE/ string : ⥸塼̾
  * @jparam	= +LIST/ string : ʸ
  * @jparam	= -STRING/ bot : ̾ 
  * @jnote	= ̾ƤʤΤξ硢Ԥޤ 
  * @jexample	= 
  > ?- strtotype("assert/1", X).
  X: assert/1 
  * @end_jexample
*/
bool strtotype_3( machine&, FSP arg1, FSP arg2, FSP arg3 )
{
//cout << "ENTER STRTOTYPE" << endl;
//	cout << arg1.DisplayLiLFeS();
	if( !arg1.IsString() )
	{
		RUNWARN("strtotype/3 requires a module name in the first argument");
//		cout << arg1.DisplayLiLFeS();
		return false;
	}
	
	if( !arg2.IsString() )
	{
		RUNWARN("strtotype/3 requires a type name in the first argument");
//		cout << arg1.DisplayLiLFeS();
		return false;
	}
	module* mod = module::SearchModule(arg1.ReadString());
	if (mod == 0) {
		RUNWARN("Warning: module "<<arg1.ReadString() << " have not loaded yet.");
		return false;
	}
	
	const type *t = mod->Search(arg2.ReadString());
	
	if( t == NULL )
          return false;
	else
	{
          return arg3.Coerce(t);
	}
//cout << "LEAVE STRTOTYPE" << endl;
}

LILFES_BUILTIN_PRED_OVERLOAD_3(strtotype_3, strtotype_3, strtotype);

/**
  * @predicate	= typetostr(+TYPE, -STRING) 
  * @desc	= Convert a type into an string. This gets simple name of type.
  * @param	= +TYPE/ bot : Type name 
  * @param	= -STRING/ string : String
  * @note	= The case of numeric type is not considered. typetostr(1,X) returns integer. 
  * @example	= 
  > ?- typetostr("assert/1", X).
  X: assert/1 
  * @end_example

  * @jdesc	= ʸѴޤά̾ޤ
  * @jparam	= +TYPE/ bot : ̾ 
  * @jparam	= -STRING/ string : ʸ
  * @jnote	= ʸͤϡ켫ȤȤϤߤʤޤtypetostr(1,X)  integer ֤ޤ 
  * @jexample	= 
  > ?- typetostr("assert/1", X).
  X: assert/1 
  * @end_jexample
*/
bool typetostr( machine& m, FSP arg1, FSP arg2 )
{
  return arg2.Unify(FSP(m, arg1.GetType()->GetSimpleName()));
}

LILFES_BUILTIN_PRED_2(typetostr, typetostr);

/**
  * @predicate	= feattostr(+TYPE, -STRING) 
  * @desc	= convert a feature to a string
  * @param	= +TYPE/ feature : Feature name 
  * @param	= -STRING/ string : String
  * @example	= 
  * @end_example

  * @jdesc	= ʸѴޤ
  * @jparam	= +TYPE/ feature : ̾ 
  * @jparam	= -STRING/ string : ʸ
  * @jexample	= 
  * @end_jexample
*/
bool feattostr( machine& m, FSP arg1, FSP arg2 )
{
	const feature *f = feature::Search(arg1.GetType());
	if( f == NULL || arg2.Unify(FSP(m, f->GetSimpleName())) == false )
          return false;
        else
          return true;
}

LILFES_BUILTIN_PRED_2(feattostr, feattostr);

/**
  * @predicate	= inttostr(+INT, -STRING) 
  * @desc	= Convert an integer into an string 
  * @param	= +INT/ int  : Number
  * @param	= -STRING/ string : String 
  * @example	= 
  > ?- inttostr(123, X).
  X: "123" 
  * @end_example

  * @jdesc	= ʸѴޤ 
  * @jparam	= +INT/ int : 
  * @jparam	= -STRING/ string : ʸ 
  * @jexample	= 
  > ?- inttostr(123, X).
  X: "123" 
  * @end_jexample
*/
bool inttostr( machine& m, FSP arg1, FSP arg2 ) {
  if (!arg1.IsInteger()) {
    return false;
  }
  int val = arg1.ReadInteger();
  bool Isminus;
  if (val < 0) {
    Isminus = true;
    val = -val;
  } else {
    Isminus = false;
  }
  char buf[102];
  buf[101] = '\0';
  int idx = 100;
  do {
    buf[idx--] = (val % 10) + '0';
    val /= 10;
  } while (val != 0);
  if (Isminus)
    buf[idx--] = '-';
  return arg2.Unify(FSP(m, &buf[idx+1]));
}

LILFES_BUILTIN_PRED_2(inttostr, inttostr);

/**
  * @predicate	= strtoint(+STRING, -INT) 
  * @desc	= Convert a string into an integer 
  * @param	= +STRING/ string : String 
  * @param	= -INT/ int  : Number
  * @example	= 
  > ?- inttostr(123, X).
  X: "123" 
  * @end_example

  * @jdesc	= ʸѴޤ 
  * @jparam	= +STRING/ string : ʸ 
  * @jparam	= -INT/ int : 
  * @jexample	= 
  > ?- inttostr("123", X).
  X: 123
  * @end_jexample
*/
bool strtoint( machine& m, FSP arg1, FSP arg2 ) {
  cell c = arg1.ReadCell();
  if( !IsSTG(c) ) {
    runwarn() << "strtoint/2 requires a string" << endmsg;
    return false;
  }
  char *s = c2STG(c);
  int val = 0;
  bool Isminus;
  if (*s == '-') {
    Isminus = true;
    s++;
  }
  else 
    Isminus = false;

  while( *s ) {
    val *= 10;
    if (!isdigit(*s)) {
      return false;
    }
    val += ((*s) - '0');
    s++;
  }
  if (Isminus) val = -val;
  return arg2.Unify(FSP(m, (mint)val));
}

LILFES_BUILTIN_PRED_2(strtoint, strtoint);

/**
  * @predicate	= floattostr(+FLOAT, -STRING) 
  * @desc	= Convert a float into a string.
  * @param	= +FLOAT/ float  : Float
  * @param	= -STRING/ string  : String 
  * @example	= 
  > ?- floattostr(1.23, X).
  X: "1.23000" 
  * @end_example
  * @jdesc	= float  string Ѵޤ
  * @jparam	= +FLOAT/ float  : Float
  * @jparam	= -STRING/ string  : String 
*/
bool floattostr( machine&, FSP arg1, FSP arg2 ) {
    if (!arg1.IsFloat()) {
      return false;
    }
    float val = arg1.ReadFloat();
    std::ostringstream oss;
    oss.setf(std::ios::showpoint);
    oss << val;
  return arg2.Unify(oss.str().c_str());
}

LILFES_BUILTIN_PRED_2(floattostr, floattostr);

/**
  * @predicate	= strtofloat(+STRING, -FLOAT) 
  * @desc	= Convert a float into a string.
  * @param	= +STRING/ string  : String 
  * @param	= -FLOAT/ float  : Float
  * @example	= 
  > ?- strtofloat("1.23", X).
  X: 1.23000 
  * @end_example
  * @jdesc	= string  float Ѵޤ
  * @jparam	= +STRING/ string  : String 
  * @jparam	= -FLOAT/ float  : Float
*/
bool strtofloat( machine& m, FSP arg1, FSP arg2 ) {
    cell c = arg1.ReadCell();
    if( !IsSTG(c) ) {
        RUNWARN( "strtofloat/2 requires a string" );
	return false;
    }
    char *s = c2STG(c);
    char* endptr;
    mfloat val = strtod(s, &endptr);
    //cerr << "val=" << val << ", errno=" << errno << endl;
    //cerr << "endptr=\"" << endptr << "\""<< endl;
    if ( endptr[0] != '\0' ) {   // Encountered a non-digit character
	return false;
    }
    if ( val == 0.0 && errno == EINVAL ) {   // Invalid format
	return false;
    }
  return arg2.Unify(FSP(m, (mfloat) val));
}

LILFES_BUILTIN_PRED_2(strtofloat, strtofloat);

/**
 * @predicate = chop(+$X, -$Y)
 * @param     = +$X/string
 * @param     = -$Y/string
 * @desc      = same as Perl
 * Same as 'chop' in Perl
*/
bool chop( machine&, FSP arg1, FSP arg2 ) {
  if ( ! arg1.IsString() ) {
    RUNWARN( "arg1 of chop/2 must be a string" );
    return false;
  }
  string s = arg1.ReadString();
  return arg2.Unify( s.substr( 0, s.size() - 1 ).c_str() );
}

LILFES_BUILTIN_PRED_2(chop, chop);

/**
 * @predicate = join(+$LIST, +$STR, -$STRING)
 * @param     = +$LIST/list of strings
 * @param     = +$STR/string
 * @param     = -$STRING/string
 * @desc      = same as Perl
 * Same as 'join' in Perl.
 * @example =
 * join(["He", "runs"], "__", "He__runs")
 * @end_example
*/
bool join( machine&, FSP arg1, FSP arg2, FSP arg3 ) {
  if ( ! arg2.IsString() ) {
    RUNWARN( "arg2 of join/3 must be a string" );
    return false;
  }
  if ( arg1.GetType() == nil ) {
    return arg3.Unify( "" );
  }
  if ( arg1.GetType() == cons ) {
    string c = arg2.ReadString();
    FSP head = arg1.Follow( hd );
    if ( ! head.IsString() ) {
      RUNWARN( "arg1 of join/3 must be a list of strings" );
      return false;
    }
    string s( head.ReadString() );
    arg1 = arg1.Follow( tl );
    while ( arg1.GetType() == cons ) {
      head = arg1.Follow( hd );
      if ( ! head.IsString() ) {
        RUNWARN( "arg1 of join/3 must be a list of strings" );
        return false;
      }
      s += c + head.ReadString();
      arg1 = arg1.Follow( tl );
    }
    return arg3.Unify( s.c_str() );
  }
  RUNWARN( "arg1 of join/3 must be a list of strings" );
  return false;
}

LILFES_BUILTIN_PRED_3(join, join);

/**
 * @predicate = concatenate_string(+$X, +$Y, -$Z)
 * @param     = +$X/string
 * @param     = +$Y/string
 * @param     = -$Z/string
 * @desc      = concatenation of strings
 * This lets $Z be the concatination of $X and $Y.
*/
/**
 * @predicate = concatenate_string(+$L, -$R)
 * @param     = +$L/ list of string
 * @param     = -$R/ string
 * @desc      = concatenation of strings
 * This lets $R be the concatination of all the elements in $L.
*/
bool concatenate_string( machine& m, FSP arg1, FSP arg2 ) {
  vector< string > str_list;
  if ( ! lilfes_to_c< vector< string > >().convert( m, arg1, str_list ) ) {
    RUNWARN( "arg1 of concatenate_string/2 must be a list of strings" );
    return false;
  }
  string s;
  for ( vector< string >::const_iterator it = str_list.begin();
        it != str_list.end();
        ++it ) {
    s += *it;
  }
  return arg2.Unify( s.c_str() );
}

LILFES_BUILTIN_PRED_2( concatenate_string, concatenate_string );
LILFES_BUILTIN_PRED_OVERLOAD_3( strcat, concatenate_string_3, concatenate_string );

} // namespace builtin

} // namespace lilfes

