/**
 * @module  = trace
 * @cvs     = $Id: trace.cpp,v 1.3 2004/05/14 08:18:53 yusuke Exp $
 * @copyright = Copyright (c) 1997-1998, Makino Takaki
 * @copyright = You may distribute this file under the terms of the Artistic License.
 * @desc    = Trace
 * Predicates for tracing execution of a program
 * @jdesc   = ȥ졼
 * @japanese =
 * ץμ¹Ԥȥ졼뤿νҸǤ
 * @end_japanese
*/

#include "builtin.h"
#include "errors.h"

namespace lilfes {

namespace builtin {

static void SetTrace(FSP F, bool b)
{
	int arity = MAXARGS-1;
	if( ! F.GetType()->IsSubType(pred) )
		INERR(F.GetMachine(), "Trace function called to non-predicate type" << F.GetType()->GetName());
	else
	{
		for( int i=1; i<MAXARGS; i++ )
			if( ! F.GetType()->IsSubType(t_pred[i]) )
			{
				arity = i-1;
				break;
			}
		procedure::New(F.GetType(), arity)->SetTrace(b);
	}
}

/**
  * @predicate	= trace(+$PRED) 
  * @desc	= Make progress of executing <i>$PRED</i> output.
  * @param	= +$PRED/pred or list :predicate or list of predicate
  * @note       = <i>$CPRED</i> must be specified with arity.And don't forget ":-"
  * @example	= 
  > :- trace('append/3').
  > :- trace(['append/3', 'reverse/2', '=/2']).

  This is a example after specify append/3 to TRACE. 

  > append([],X,X).
  > append([A|X], Y, [A|Z]) :- append(X,Y,Z).
  > ?- append(X,Y,[a]).
  TRACE: ENTER append/3(1)    - enter the first clause of append.
  bot, bot, < a >             - arity is (bot, bot, [a]) .
  TRACE: SUCCESS append/3(1)  - success the first clause of append.
  < >, < a >, < a >           - arity become ([], [a], [a]). 
  X: < >                      - output of result.
  Y: < a >
  Enter ';' for more choices, otherwise press ENTER --> ;
  TRACE: REENTER append/3(2)  - backtrack and reenter the second clause of append.
  bot, bot, < a >             - arity is (bot, bot, [a]) .
  TRACE: ENTER append/3(1)    --- enter the first clause of append.
  list, bot, < >              --- arity is (list, bot, []) .
  TRACE: SUCCESS append/3(1)  --- success the first clause of append.
  < >, < >, < >               --- arity is  ([], [], []). 
  TRACE: SUCCESS append/3(2)  - success the cause of call,append.
  < a >, < >, < a >           - arity is ([a], [], [a]).
  X: < a >                    - output of result.
  Y: < >
  Enter ';' for more choices, otherwise press ENTER --> ;
  TRACE: REENTER append/3(2)  --- backtrack append that called in front. 
  list, bot, < >              --- arity is  (list, bot, []).
  TRACE: FAIL append/3        --- fail that append.
  list, bot, < >              --- arity is (list, bot, []) .
  TRACE: FAIL append/3        - fail the cause of call, append
  bot, bot, < a >             - arity is (bot, bot, [a]).
  no                          - return no.
  * @end_example

  * @jdesc		= Ҹ<i>$PRED</i>μ¹Էвɽ褦ˤޤ 
  * @jparam		= +$PRED/pred or list :Ҹ ޤϤΥꥹ
  * @jnote      =  Ҹ̾arityդǻꤷޤޤ:- ˺ʤ褦ˤƲ  
  * @jexample	=
  > :- trace('append/3').
  > :- trace(['append/3', 'reverse/2', '=/2']).

  append/3 TRACEꤷμ¹Ǥ
  > append([],X,X).
  > append([A|X], Y, [A|Z]) :- append(X,Y,Z).
  > ?- append(X,Y,[a]).
  TRACE: ENTER append/3(1)    - append 1ܤޤ
  bot, bot, < a >             -  (bot, bot, [a]) Ǥ
  TRACE: SUCCESS append/3(1)  - append 1ܤ᤬ޤ
  < >, < a >, < a >           -  ([], [a], [a]) ˤʤޤ
  X: < >                      - ɽǤ
  Y: < a >
  Enter ';' for more choices, otherwise press ENTER --> ;
  TRACE: REENTER append/3(2)  - Хåȥå append 2ܤޤ
  bot, bot, < a >             -  (bot, bot, [a]) Ǥ
  TRACE: ENTER append/3(1)    --- append 1ܤ᤬ƤӽФޤ
  list, bot, < >              ---  (list, bot, []) Ǥ
  TRACE: SUCCESS append/3(1)  --- append 1ܤ᤬ޤ
  < >, < >, < >               ---  ([], [], []) Ǥ
  TRACE: SUCCESS append/3(2)  - ƤӽФ append ޤ
  < a >, < >, < a >           -  ([a], [], [a]) Ǥ
  X: < a >                    - ɽǤ
  Y: < >
  Enter ';' for more choices, otherwise press ENTER --> ;
  TRACE: REENTER append/3(2)  --- ǸƤФ줿 append ˥Хåȥåޤ
  list, bot, < >              ---  (list, bot, []) Ǥ
  TRACE: FAIL append/3        ---  append Ԥޤ
  list, bot, < >              ---  (list, bot, []) Ǥ
  TRACE: FAIL append/3        - ƤӽФ append Ԥޤ
  bot, bot, < a >             -  (bot, bot, [a]) Ǥ
  no                          - no ֤ޤ
  
  * @end_jexample
*/

  bool trace( machine&, FSP arg1 )
  {
    while( arg1.GetType()->IsSubType(cons) )
      {
        SetTrace(arg1.Follow(hd), true);
        arg1 = arg1.Follow(tl);
      }
    if( ! arg1.GetType()->IsSubType(nil) )
      SetTrace(arg1, true);
    return true;
  }

/**
  * @predicate	= notrace(+$PRED) 
  * @desc	= Disable tracing of <i>$PRED</i>.
  * @param	= +$PRED/pred or list :predicate or list of predicate
  * @note       = <i>$CPRED</i> must be specified with arity.
  * @example	= 
  > :- notrace('append/3').
  > :- notrace(['append/3', 'reverse/2', '=/2']).
  * @end_example
  * @jdesc		= Ҹ<i>$PRED</i>Υȥ졼ޤ
  * @jparam		= +$PRED/pred or list :Ҹ ޤϤΥꥹ
  * @jnote      =  Ҹ̾arityդǻꤷޤ
*/

  bool notrace( machine&, FSP arg1 )
  {
    while( arg1.GetType()->IsSubType(cons) )
      {
        SetTrace(arg1.Follow(hd), false);
        arg1 = arg1.Follow(tl);
      }
    if( ! arg1.GetType()->IsSubType(nil) )
      SetTrace(arg1, false);
    return true;
  }

  LILFES_BUILTIN_PRED_1(trace, trace);
  LILFES_BUILTIN_PRED_1(notrace, notrace);

}

} // namespace lilfes
