#include "utility.h"
#include <sstream>

namespace lilfes {

using namespace std;

// e.g. for 'append/3'
procedure*
prepare_proc(const type* t)
{
  if ( t == NULL ) return NULL;
  procedure* p = procedure::Search( t );
  return p;
}

// e.g. for 'append'
procedure*
prepare_proc(const type* t, size_t nargs)
{
  // original version
  /*
  ostringstream pname;
  pname << name << '/' << nargs;

  const type *t = mod->Search(pname.str().c_str());
  if (!t) {
    return NULL;
  }
  procedure* p = procedure::Search(t);
  if (!p) {
    return NULL;
  }
  int nf = t->GetNFeatures();
  assert(nf == nargs);

  return p;
  */

  // new version
  if ( t == NULL ) return NULL;
  t = t->TypeUnify( t_pred[ nargs ] );
  procedure* p = prepare_proc( t );
  if ( p && p->GetArity() == nargs ) return p;
  return NULL;
}

// e.g. for "append"
procedure*
prepare_proc(const module* mod, const string & name, size_t nargs)
{
  const type* t = mod->Search( name.c_str() );
  return prepare_proc( t, nargs );
}

bool
call_proc(machine& mach,
          procedure* proc,
          vector<FSP>& args) {
  size_t nargs = args.size();
  procthunk *pt = proc->GetThunk();
  assert(pt != NULL);

  mach.SetRP(nargs);
  for (size_t i = 0; i < nargs; i++) {
    mach.WriteLocal(i, PTR2c(args[i].GetAddress()));
  }

  core_p trailp = mach.SetTrailPoint( NULL );
  pt->exec(mach);

  if( mach.GetIP() == NULL ||
      ( ! mach.Execute(mach.GetIP()) ) ) {
    mach.TrailBack( trailp );
    return false;
  }
  return true;
}

void
findall_proc(machine& mach,
             procedure* proc,
             const vector<FSP> & args,
             const vector<bool> & output_flags,
             list<vector<FSP> > & results)
{
  int nargs = args.size();
  procthunk *pt =proc->GetThunk();
  assert(pt != NULL);
        
  mach.SetRP(nargs);

  for (int i = 0; i < nargs; i++) {
    mach.WriteLocal(i, PTR2c(args[i].GetAddress()));
  }

  int oldheap = mach.GetCurrentSlot();
  core_p save = mach.SwitchNewHeap();
  int newheap = mach.GetCurrentSlot();

  code *oldIP = mach.GetIP();
  core_p TrailP = mach.SetTrailPoint(NULL);

  pt->exec(mach);

  results.clear();
  bool loop = false;
  if( mach.GetIP() != NULL )
    do {
      loop = loop==false ? mach.Execute(mach.GetIP())   // The first answer
        : mach.NextAnswer(); // Quest for the next answer

      if( loop )
        {
          mach.GoToHeap(oldheap);
          vector<FSP> v;
          for (int i = 0; i < nargs; i++) {
            if (output_flags[i]) {
              FSP p(args[i].Copy());
              v.push_back(p);
              //              cerr << v.size() << " " << (int)p << endl;
            }
          }
          results.push_back(v);

          mach.GoToHeap(newheap);
        }
    } while( loop );

  mach.TrailBack(TrailP);
  mach.SwitchOldHeap(save);
  mach.SetIP(oldIP);
}

//////////////////////////////////////////////////////////////////////

module* load_module( machine& mach, const std::string& title, const std::string& module_name, module* parent_module ) {
  core_p trailp = mach.SetTrailPoint( NULL );
  istringstream dummy_stream("");
  lexer_t tmp_lex(&dummy_stream, "stdin");
  lexer_t old_lex = mach.getLexer();
  mach.setLexer(tmp_lex);
  module* mod = module::new_module( mach, title.c_str(), module_name.c_str(), module::NMT_ENSURE_INITIAL, parent_module );
  if ( mod ) {
    mach.parse();
  }
  mach.setLexer(old_lex);
  mach.TrailBack( trailp );
  return mod;
}

//////////////////////////////////////////////////////////////////////

// template <>
// bool list_to_vector< FSP >( machine&,
//                             FSP fsp,
//                             std::vector< FSP >& vec ) {
//   while ( fsp.GetType() == cons ) {
//     FSP head = fsp.Follow( hd );
//     vec.push_back( head );
//     fsp = fsp.Follow( tl );
//   }
//   return true;
// }

// template <>
// bool list_to_vector< std::string >( machine&,
//                                     FSP fsp,
//                                     std::vector< std::string >& vec ) {
//   while ( fsp.GetType() == cons ) {
//     FSP head = fsp.Follow( hd );
//     if ( ! head.IsString() ) return false;
//     vec.push_back( head.ReadString() );
//     fsp = fsp.Follow( tl );
//   }
//   return true;
// }

// template <>
// bool list_to_vector< int >( machine&,
//                             FSP fsp,
//                             std::vector< int >& vec ) {
//   while ( fsp.GetType() == cons ) {
//     FSP head = fsp.Follow( hd );
//     if ( ! head.IsInteger() ) return false;
//     vec.push_back( head.ReadInteger() );
//     fsp = fsp.Follow( tl );
//   }
//   return true;
// }

// template <>
// bool list_to_vector< float >( machine&,
//                               FSP fsp,
//                               std::vector< float >& vec ) {
//   while ( fsp.GetType() == cons ) {
//     FSP head = fsp.Follow( hd );
//     if ( ! head.IsFloat() ) return false;
//     vec.push_back( head.ReadFloat() );
//     fsp = fsp.Follow( tl );
//   }
//   return true;
// }

} // namespace lilfes

